//-*****************************************************************************
//
// Copyright (c) 2009-2011,
//  Sony Pictures Imageworks, Inc. and
//  Industrial Light & Magic, a division of Lucasfilm Entertainment Company Ltd.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
// *       Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// *       Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
// *       Neither the name of Sony Pictures Imageworks, nor
// Industrial Light & Magic nor the names of their contributors may be used
// to endorse or promote products derived from this software without specific
// prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-*****************************************************************************

#include <Alembic/Abc/OArchive.h>
#include <Alembic/Abc/OObject.h>

namespace Alembic {
namespace Abc {
namespace ALEMBIC_VERSION_NS {

//-*****************************************************************************
// This is mostly here to provide debug support when tracing down
// bad reference counting.
OArchive::~OArchive()
{
    //std::cout << "OArchive::~OArchive() name: "
    //          << this->getName() << std::endl
    //          << "\tTop Object use count: " << m_topObject.use_count()
    //          << std::endl;
}


//-*****************************************************************************
std::string OArchive::getName() const
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::getName()" );

    return m_archive->getName();

    ALEMBIC_ABC_SAFE_CALL_END();

    // Not all error handlers throw,
    // so return a NO-OP value
    return "";
}

//-*****************************************************************************
int8_t OArchive::getCompressionHint() const
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::getCompressionHint" );

    return m_archive->getCompressionHint();

    ALEMBIC_ABC_SAFE_CALL_END();

    // Not all error handlers throw,
    // so return a NO-OP value
    return -1;
}

//-*****************************************************************************
void OArchive::setCompressionHint( int8_t iCh )
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::setCompressionHint" );

    m_archive->setCompressionHint( iCh );

    ALEMBIC_ABC_SAFE_CALL_END();
}

//-*****************************************************************************
uint32_t OArchive::addTimeSampling( const AbcA::TimeSampling & iTs )
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::addTimeSampling" );

    return m_archive->addTimeSampling( iTs );

    ALEMBIC_ABC_SAFE_CALL_END();

    // Not all error handlers throw,
    // so return a NO-OP value
    return 0;
}

//-*****************************************************************************
AbcA::TimeSamplingPtr OArchive::getTimeSampling( uint32_t iIndex )
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::getTimeSampling" );

    return m_archive->getTimeSampling( iIndex );

    ALEMBIC_ABC_SAFE_CALL_END();

    // Not all error handlers throw,
    // so return a NO-OP value
    return AbcA::TimeSamplingPtr();
}

//-*****************************************************************************
uint32_t OArchive::getNumTimeSamplings()
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::getNumTimeSampling" );

    return m_archive->getNumTimeSamplings( );

    ALEMBIC_ABC_SAFE_CALL_END();

    // Not all error handlers throw,
    // so return a NO-OP value
    return 0;
}

//-*****************************************************************************
OObject OArchive::getTop()
{
    ALEMBIC_ABC_SAFE_CALL_BEGIN( "OArchive::getTop()" );

    return OObject( m_archive->getTop() );

    ALEMBIC_ABC_SAFE_CALL_END();

    // Not all error handlers throw, so here is a default behavior.
    return OObject();
}

} // End namespace ALEMBIC_VERSION_NS
} // End namespace Abc
} // End namespace Alembic
